Attribute VB_Name = "ErrorMod"
' '---------------------------------------------------------------------------------------
' Name     : ErrorMod
' Purpose  : Minimal error-handling module to avoid compile errors in sample code from nolongerset.com.
' Author   : Mike Wolfe
' Date     : 12/4/2023
' '---------------------------------------------------------------------------------------
Option Explicit
Option Compare Database

'Error logging options
Public Enum elOptions
    elHide = &H0
    elShow = &H1
    elReraise = &H2
End Enum

Private Const mDO_NOT_REPORT_ERR_NUM As Long = 65535

'---------------------------------------------------------------------------------------
' Procedure : Throw
' Author    : Mike Wolfe
' Source    : https://nolongerset.com/throwing-errors-in-vba/
' Date      : 7/24/2014 - 12/7/2016
' Purpose   : Throws a new error, calculating a unique error number by hashing the Msg.
' Notes     - Msg may contain one or more placeholders which will be substituted
'               with contents of the Notes() array.
'           - Contents of the Notes array will be enclosed in single quotes when they are
'               added to show that they are custom for a particular instance.
'           - The placeholders may contain quotes around them but it is not necessary.
'           - Tabs and newlines can be passed using escape sequences, \t and \n
' Usage     : Throw "Could not find file {0} for account:\n\n{1}", FName, AcctNum
'   Results : Could not find file 'C:\Temp\MyFile.txt' for account:
'
'             '123456'
'---------------------------------------------------------------------------------------
'
Public Sub Throw(Msg As String, ParamArray Notes() As Variant)
    Dim s As String, i As Integer
    s = Msg
    s = Replace(s, "\t", vbTab)
    s = Replace(s, "\n", vbNewLine)
    For i = LBound(Notes()) To UBound(Notes())
        Dim ReplTxt As String
        ReplTxt = "'" & Notes(i) & "'"
        s = Replace(s, "'{" & i & "}'", ReplTxt)
        s = Replace(s, "{" & i & "}", ReplTxt)
    Next i
    Err.Raise ErrNumFromMsg(Msg), , s
End Sub

'---------------------------------------------------------------------------------------
' Procedure : Alert
' Author    : Mike
' Date      : 12/29/2015 - 8/18/2017
' Purpose   : Throws a new error using a reserved error number to prevent FogBugz reporting.
' Notes     - This is exactly the same as the above Throw() procedure, but it will not
'               generate a FogBugz report.
'           - Use this when you want to inform the user of an error and have it bubble up the
'               call stack, but you're not interested in seeing it show up in a FogBugz error report.
'           - These errors will still be logged to text files or local databases if that
'               is set up; only the FogBugz reporting is affected.
'---------------------------------------------------------------------------------------
'
Public Sub Alert(Msg As String, ParamArray Notes() As Variant)
    Dim s As String, i As Integer
    s = Msg
    s = Replace(s, "\t", vbTab)
    s = Replace(s, "\n", vbNewLine)
    For i = LBound(Notes()) To UBound(Notes())
        Dim ReplTxt As String
        ReplTxt = "'" & Notes(i) & "'"
        s = Replace(s, "'{" & i & "}'", ReplTxt)
        s = Replace(s, "{" & i & "}", ReplTxt)
    Next i
    
    '--== Uncomment for use with vbWatchdog ==--
    'With ErrEx.LiveCallstack
    '    .FirstLevel
    '    If .NextLevel Then
    '        mAlertCaller = " (" & .ModuleName & "." & .ProcedureName & ")"
    '    Else
    '        mAlertCaller = vbNullString
    '    End If
    'End With
    '---------------------------------------------
    
    Err.Raise mDO_NOT_REPORT_ERR_NUM, , s
End Sub

'---------------------------------------------------------------------------------------
' Procedure : ErrNumFromMsg
' Author    : Mike
' Date      : 9/22/2015
' Purpose   : Converts a msg into a somewhat-unique hash.  This is a public convenience
'               function primarily made available for testing.
'---------------------------------------------------------------------------------------
'
Public Function ErrNumFromMsg(Msg As String) As Long
    ErrNumFromMsg = vbObjectError + GenerateErrNumFromMsgHash(Msg)
End Function

'---------------------------------------------------------------------------------------
' Procedure : GenerateErrNumFromMsgHash
' Author    : Mike
' Date      : 7/24/2014
' Purpose   : Given a string of text, reliably reproduces a number between 51365535
'               to use when generating user-defined errors.
' Notes     - This is an overly simplistic hash method.  It is likely rife with potential
'               collisions.  However, we are not too concerned with the occasional
'               collision, so it seems like a good trade-off.
' Usage     :
'>> GenerateErrNumFromMsgHash("Could not find file {0} for account {1}")
' 41945
'---------------------------------------------------------------------------------------
'
Private Function GenerateErrNumFromMsgHash(Msg As String) As Long
    Const ValRange As Long = 65535 - 513
    Const ShiftVal As Long = 513
    Dim i As Long, Val As Long
    For i = 1 To Len(Msg)
        Val = (Val + (Asc(Mid(Msg, i, 1)) ^ 2)) Mod ValRange
    Next i
    GenerateErrNumFromMsgHash = Val + ShiftVal
    If GenerateErrNumFromMsgHash = mDO_NOT_REPORT_ERR_NUM Then
        'make sure we don't accidentally prevent a reportable message from being reported
        GenerateErrNumFromMsgHash = mDO_NOT_REPORT_ERR_NUM - 1
    End If
 End Function

' ----------------------------------------------------------------
' Procedure : LogErr
' Date      : 12/4/2023
' Author    : Mike Wolfe
' Source    : https://nolongerset.com/
' Purpose   : Placeholder for a more robust error logging function.
' Notes     - Many of my code samples include calls to my LogErr()
'               procedure, which is tightly integrated with vbWatchdog.
'               This procedure is a minimal implementation of an error
'               handler to allow those code samples to compile.  You
'               should replace this routine with your own custom error
'               handler when implementing my sample code into your own
'               projects.
' ----------------------------------------------------------------
Sub LogErr(ErrObj As ErrObject, Errs As Errors, ModName As String, ProcName As String, _
           Optional Options As elOptions = elShow, Optional Notes As Variant = Null)
    Dim Msg As String
    Msg = "Error in " & ModName & "." & ProcName & vbNewLine & vbNewLine & _
          ErrObj.Description + vbNewLine & Notes
    
    If (Options And elShow = elShow) Then
        MsgBox Msg, vbExclamation, "Error: " & ErrObj.Number
    Else
        Debug.Print "Error: " & ErrObj.Number, Msg
    End If
End Sub

